<html>
    <head>
        <title>{{page_title}}</title>
    </head>
    <body>
            <h1>{{title}}</h1>
            <p>{{text}}</p>
            <a href={{dir}}>
                See Snekmer clustering results in file explorer: (<b>output/cluster/</b>)
            </a>
            <br>
            <a href={{table}}>
                Direct link to results table: (<b>snekmer.csv</b>)
            </a>
            <h2>{{image1_name}}</h2>
            <img src={{image1_path}} width="400" height="300" />
            <h2>{{image2_name}}</h2>
            <img src={{image2_path}} width="400" height="300" />
            <h2>{{image3_name}}</h2>
            <img src={{image3_path}} width="400" height="300" />
    </body>
</html>